#ifndef DIFFERENTIAL_DISTRIBUTIONS_H
#define DIFFERENTIAL_DISTRIBUTIONS_H

#include "sm_control.h"
#include "utils.h"

namespace CuTe {

namespace CuTe_MC {

namespace NS_diff_distris {

using Shared_Memory_Control::SM_Array;
using Shared_Memory_Control::Mutex;
using Shared_Memory_Control::SM_Variable;

using NS_Utils::NS_pow::pow_2;

using std::endl;
using std::ofstream;
using std::string;
using std::stringstream;
using std::vector;

typedef double val_T;
typedef long long int cnt_T;
typedef double(*Func_of_coords)(const vector<double>& coords, void* params);

class Differential_Distribution{
public:
	string name;
protected:
	Func_of_coords x_of_coords;
	void*	params;

	//Distribution of N bins in [min,max]
	double min;
	double max;
	int N;

	double dx;	//	= (max-min)/N
	double xd;	//	= 1./dx

	int ncores;
	bool attached;
	mutable Mutex mutex;
	SM_Variable<int> SM_merge_count;

	//Local Variables for every bin i
	//addval(..) -> Sum up every value i for one iteration a
	//merge_me() resets them
	vector<cnt_T> nai;
	vector<val_T> wvai;
	vector<val_T> wfai;
	vector<val_T> Svai;
	vector<val_T> Sfai;

	//Shared variables
	//merge_me() -> Sum of the results above over every i and core
	//merge_all() resets them
	SM_Variable<cnt_T> SM_na;

	//Shared variables
	//merge_me() -> Sum of the results above over every core
	//merge_all() resets them
	SM_Array<cnt_T> SM_nai;
	SM_Array<val_T> SM_wvai;
	SM_Array<val_T> SM_wfai;
	SM_Array<val_T> SM_Svai;
	SM_Array<val_T> SM_Sfai;

	//Shared variables
	//merge_all() -> Sum of the results above over every iteration a
	SM_Array<cnt_T> SM_ni;
	SM_Array<val_T> SM_Wvi;
	SM_Array<val_T> SM_Vi;
	SM_Array<val_T> SM_Wfi;
	SM_Array<val_T> SM_Fi;

	//Shared variables
	//merge_all() -> Sum of the results above over every i and a
	SM_Variable<cnt_T> SM_n;


public:

	inline int merge_me(int SID,int iteration){
		NS_Utils::DOUT dout("Differential_Distribution::merge_me",name);
		dout()	<<"SID="<<SID	<<"\t"<<"iteration="<<iteration	<<endl;

		vector<cnt_T>::iterator	ina		= nai.begin();
		vector<val_T>::iterator	iwva	= wvai.begin();
		vector<val_T>::iterator	iSva 	= Svai.begin();
		vector<val_T>::iterator	iwfa	= wfai.begin();
		vector<val_T>::iterator	iSfa	= Sfai.begin();
		cnt_T* iSM_na		= SM_nai.begin();
		val_T* iSM_wva		= SM_wvai.begin();
		val_T* iSM_Sva		= SM_Svai.begin();
		val_T* iSM_wfa		= SM_wfai.begin();
		val_T* iSM_Sfa		= SM_Sfai.begin();
		for(int i=0;i<N;++i){
			if(*ina){
				(*iSM_wva) +=(*iwva);
				(*iSM_Sva) +=(*iSva);
				(*iSM_wfa) +=(*iwfa);
				(*iSM_Sfa) +=(*iSfa);
				if(*iSM_na){
					val_T n		= static_cast<val_T>( (*iSM_na)+(*ina) );
					val_T tmp	= static_cast<val_T>(*ina)/static_cast<val_T>(n*(*iSM_na));
					n  /= static_cast<val_T>(*ina);
					(*iSM_Sva)	+= tmp*pow_2(n*(*iwva)-(*iSM_wva));
					(*iSM_Sfa)	+= tmp*pow_2(n*(*iwfa)-(*iSM_wfa));
				}
				(*iSM_na) +=(*ina);

				(*SM_na)  +=(*ina);

			}
			(*(ina++))=0.;
			(*(iwva++))=0.;
			(*(iSva++))=0.;
			(*(iwfa++))=0.;
			(*(iSfa++))=0.;
			++iSM_na;
			++iSM_wva;
			++iSM_Sva;
			++iSM_wfa;
			++iSM_Sfa;
		}

		return (++(*SM_merge_count));
	}

	inline int merge_all(int SID,int iteration){
		NS_Utils::DOUT dout("Differential_Distribution::merge_all",name);
		dout()	<<"SID="<<SID	<<"\t"<<"iteration="<<iteration	<<endl;

		int _tmp = (*SM_merge_count)%ncores;
		if(_tmp) return _tmp;	// Not ready for merge_all

		if((*SM_merge_count)==0) return 0;	// already merged

		val_T n = static_cast<val_T>(*SM_na);
		(*SM_n)	+=(*SM_na);
		(*SM_na) = 0;


		cnt_T* ina		= SM_nai.begin();
		val_T* iwva		= SM_wvai.begin();
		val_T* iSva		= SM_Svai.begin();
		val_T* iwfa		= SM_wfai.begin();
		val_T* iSfa		= SM_Sfai.begin();
		cnt_T* in		= SM_ni.begin();
		val_T* iWv		= SM_Wvi.begin();
		val_T* iV		= SM_Vi.begin();
		val_T* iWf		= SM_Wfi.begin();
		val_T* iF		= SM_Fi.begin();
		for(int i=0;i<N;++i){
			if(*ina){
				val_T	na	= static_cast<val_T>(*ina);
				val_T tmpWv = 1.;
				val_T tmpWf = 1.;
				if(*ina>1){
					(*in) 	   += (*ina);
					val_T tmp2	= na*(n-1.);

					tmpWv = tmp2/( n*na*(*iSva) + (n-na)*pow_2(*iwva) );
					tmpWf = tmp2/( n*na*(*iSfa) + (n-na)*pow_2(*iwfa) );
#ifndef OLD_WEIGHT_SUM
					tmpWv*=pow_2(*iwva);
					tmpWf*=pow_2(*iwfa);
#endif
#ifdef OLD_WEIGHT_SUM
					(*iWv) += tmpWv;
					(*iV)  += tmpWv*(*iwva);
					(*iWf) += tmpWf;
					(*iF)  += tmpWf*(*iwfa);
#endif
				}
#ifdef OLD_WEIGHT_SUM
				else{
					tmpWv /= fabs(*iwva);
					tmpWf /= fabs(*iwfa);

					(*iV)  += tmpWv;
					(*iF)  += tmpWf;

					(*iWv) += pow_2(tmpWv);
					(*iWf) += pow_2(tmpWf);
				}
#endif
#ifndef OLD_WEIGHT_SUM
				(*iWv) += tmpWv;
				(*iV)  += tmpWv*(*iwva);
				(*iWf) += tmpWf;
				(*iF)  += tmpWf*(*iwfa);
#endif
			}
			(*(ina++))=0.;
			(*(iwva++))=0.;
			(*(iSva++))=0.;
			(*(iwfa++))=0.;
			(*(iSfa++))=0.;
			++in;
			++iWv;
			++iV;
			++iWf;
			++iF;
		}


		(*SM_merge_count)=0;
		return 0;
	}

	inline int get_i(double x)const{
		return static_cast<int>( (x-min)*xd );
	}

	inline void add_val(double x, double fx, double wx=1., double vx=1.){
		if(!attached) return;

		int i = get_i(x);
		if(i<0 || i>=N) return;

		cnt_T* n = &(nai[i]);
		wvai[i]	+= (wx*=vx);
		wfai[i] += (wx*fx);
		if((*n)++){
			val_T tmp = 1./static_cast<val_T>((*n)*((*n)-1.));
			Svai[i]	+= tmp*pow_2(wvai[i]-(wx*=static_cast<val_T>(*n)));
			Sfai[i]	+= tmp*pow_2(wfai[i]-wx*fx);
		}
	}

	inline void add_val(const vector<double>& coords, double fx, double wx=1., double vx=1.){
		return add_val(x_of_coords(coords,params),fx,wx,vx);
	}

	inline void to_File(string Filename){
		NS_Utils::DOUT dout("Differential_Distribution::to_File",name);
		if(!attached) return;

		stringstream ss;
		ss	<<Filename<<"_"<<name<<".out";
		Filename = ss.str();

		ofstream of;
		of.open(Filename.c_str());

		of	<< "#"
				<< "x\t"
				<< "f(x)\t"
				<< "dx\t"
				<< "df\t"
				<< "V\t"
				<< "dV\t"
				<< "n\t"
				<< endl;

		double x = min+0.5*dx;

		cnt_T* in		= SM_ni.begin();
		val_T* iWv		= SM_Wvi.begin();
		val_T* iV		= SM_Vi.begin();
		val_T* iWf		= SM_Wfi.begin();
		val_T* iF		= SM_Fi.begin();
		val_T Sv =0.;
		val_T  V =0.;
		val_T Sf =0.;
		val_T  F =0.;
		for(int i=0; i<N;++i){
			if(*in){
#ifndef OLD_WEIGHT_SUM
				(*iWv)=1./(*iWv);
				(*iWf)=1./(*iWf);

				V += ( (*iV)*=(*iWv) );
				F += ( (*iF)*=(*iWf) );

				Sv += ( (*iWv)*=pow_2(*iV) );
				Sf += ( (*iWf)*=pow_2(*iF) );

				of	<< x << "\t"
						<< (*iF)*xd << "\t"
						<< 0.5*dx << "\t"
						<< sqrt(*iWf)*xd << "\t"
						<< (*iV)*xd << "\t"
						<< sqrt(*iWv)*xd << "\t"
						<< *in << "\t"
						<< endl;
#endif
#ifdef OLD_WEIGHT_SUM
				(*iWv)=1./(*iWv);
				(*iWf)=1./(*iWf);

				Sv += (*iWv);
				V += (*iWv)*(*iV);
				Sf += (*iWf);
				F += (*iWf)*(*iF);

				of	<< x << "\t"
						<< (*iWf)*(*iF)*xd << "\t"
						<< 0.5*dx << "\t"
						<< sqrt(*iWf)*xd << "\t"
						<< (*iWv)*(*iV)*xd << "\t"
						<< sqrt(*iWv)*xd << "\t"
						<< *in << "\t"
						<< endl;
#endif
			}
			x+=dx;
			++in;
			++iWv;
			++iWf;
			++iV;
			++iF;
		}
		of.close();
		dout()	<< Filename <<"\tn="<<*SM_n	<<endl;//<<"\tnc="<<*SM_n_contribute
		dout()	<<"SV="<<V <<" +- "<<sqrt(Sv)	<<endl;
		dout()	<<"SF="<<F <<" +- "<<sqrt(Sf)	<<endl;
	}

	inline void reinit(){
		NS_Utils::DOUT dout("Differential_Distribution::reinit",name);
		(*SM_merge_count)=0;

		nai.assign(N,0.);
		Svai.assign(N,0.);
		wvai.assign(N,0.);
		wfai.assign(N,0.);
		Sfai.assign(N,0.);

		(*SM_na)=0.;

		SM_nai.init(N,0.);
		SM_Svai.init(N,0.);
		SM_wvai.init(N,0.);
		SM_wfai.init(N,0.);
		SM_Sfai.init(N,0.);

		SM_ni.init(N,0.);
		SM_Wvi.init(N,0.);
		SM_Wfi.init(N,0.);
		SM_Vi.init(N,0.);
		SM_Fi.init(N,0.);

		(*SM_n)=0;

		attached = true;
	}

	inline void init(double min, double max, int N, int ncores){
		NS_Utils::DOUT dout("Differential_Distribution::init",name);
		dout()	<<"min="<<min	<<"\t"<<"max="<<max	<<"\t"<<"N="<<N	<<"\t"<<"ncores="<<ncores	<<endl;

		if(N){
			this->min	= min;
			this->max	= max;
			this->N		= N;
			this->ncores= ncores;
			dx = (max-min)/static_cast<double>(N);
			xd = 1./dx;
			reinit();
		}else{
			attached=false;
		}
	}

protected:
	inline Differential_Distribution()
	:name(),x_of_coords(NULL),params(NULL)
	,min(0.),max(0.),N(0)
	,dx(0.),xd(0.)
	,ncores(-1)
	,attached(false)
	{
		NS_Utils::DOUT dout("Differential_Distribution");
		SM_n.status("SM_n");
	}

public:
	inline Differential_Distribution(const string& name, Func_of_coords x_of_coords, void* params)
	:name(name),x_of_coords(x_of_coords),params(params)
	,min(0.),max(0.),N(0)
	,dx(0.),xd(0.)
	,ncores(-1)
	,attached(false)
	{
		NS_Utils::DOUT dout("Differential_Distribution",name);
		SM_n.status("SM_n");
	}

	inline double get_x_of_coords(const vector<double>& coords)const{
		return x_of_coords(coords,params);
	}


	inline ~Differential_Distribution(){
		NS_Utils::DOUT dout("~Differential_Distribution",name);
	}

	inline void lock(){return mutex.lock();}
	inline bool try_lock(){return mutex.try_lock();}
	inline void unlock(){return mutex.unlock();}
};


class Differential_Distribution_Set{
protected:
	vector<Differential_Distribution*> dists;
	void* params;
	vector<Differential_Distribution*>::iterator it;
	vector<Differential_Distribution*>::iterator end;


	int ncores;
	int iteration;

public:
	inline void clear(){
		it	= dists.begin();
		while(it!=end){
			delete (*(it++));	// No = NULL
		}
		dists.clear();
		end				= dists.end();
	}

	inline Differential_Distribution_Set()
	:dists()
	,params(NULL)
	,it(),end(dists.end())
	,ncores(-1),iteration(-1){

	}

	inline ~Differential_Distribution_Set(){
		clear();
	}

	inline void init(int ncores,void* params){
		clear();
		this->ncores	= ncores;
		this->params	= params;
		iteration		= 1;
	}

	inline void add_dist(const string& name, Func_of_coords x_of_coords, double min, double max, int N){
		if(N>0){
			Differential_Distribution* dist = new Differential_Distribution(name,x_of_coords,params);
			dist->init(min,max,N,ncores);
			dists.push_back(dist);
			end	= dists.end();
		}
	}


	inline void merge(int SID){
		int start = SID-1;
		int ndists= dists.size();


		// List of dists i_have_to_merge
		vector<bool> i_have_to_merge(ndists,true);
		// # of dists i_have_to_merge
		int i_have_still_to_merge= ndists;


		// Copy for the second step
		vector<bool> we_have_to_merge(i_have_to_merge);
		int we_have_still_to_merge = i_have_still_to_merge;


		int wcount=0;
		while(i_have_still_to_merge>1){

			for(int i=0; i<ndists; ++i){
				if(i_have_to_merge[i] && dists[i]->try_lock()){
					dists[i]->merge_me(SID,iteration);
					dists[i]->unlock();
					i_have_to_merge[i]=false;
					i_have_still_to_merge--;
					wcount=0;
				}
			}

			if(++wcount>ndists) break;
		}

		for(int i=0; i<ndists; ++i){
			if(i_have_to_merge[i]){
				dists[i]->lock();
				dists[i]->merge_me(SID,iteration);
				dists[i]->unlock();
				//				i_have_to_merge[i]=false;
				//				i_have_still_to_merge--;
			}
		}


		wcount=0;
		while(we_have_still_to_merge>1){
			//			cout << "["<<SID<<"]("<<iteration<<")["<<wcount<<"] we_have_still_to_merge="<< we_have_still_to_merge << endl;
			for(int i=0; i<ndists; ++i){
				if(we_have_to_merge[i] && dists[i]->try_lock()){
					int tmp = dists[i]->merge_all(SID,iteration);
					dists[i]->unlock();
					if(tmp==0){
						we_have_to_merge[i]=false;
						we_have_still_to_merge--;
						wcount=0;
					}
				}
			}
			if(++wcount>ndists) break;
		}
		//		while(we_have_still_to_merge){
		for(int i=0; i<ndists; ++i){
			if(we_have_to_merge[i]){
				dists[i]->lock();
				//					int tmp =
				dists[i]->merge_all(SID,iteration);
				dists[i]->unlock();
				//					if(tmp==0){
				//						we_have_to_merge[i]=false;
				//						we_have_still_to_merge--;
				//					}
			}
		}
		//		}
		//		cout << "["<<SID<<"]("<<iteration<<")<-merge" << endl;
	}

	inline void add_value(int SID, const vector<double>& coords, int iter, double fx, double wx=1., double vx=1.){
		if(iter!=iteration){
			merge(SID);
			iteration=iter;
		}

		for(it=dists.begin(); it!=end; ++it){
			(*it)->add_val(coords,fx,wx,vx);
		}
	}

	inline void to_File(const string& filename){
		for(it=dists.begin(); it!=end; ++it){
			(*it)->to_File(filename);
		}
	}

};

}  // namespace NS_diff_distris


}  // namespace CuTe_MC


}  // namespace CuTe

#endif
